"""
ttldict2
Another dictionary with expiring keys.

:copyright: 2021 Nándor Mátravölgyi
:license: Apache2, see LICENSE for more details.
"""

__author__ = "Nándor Mátravölgyi"
__copyright__ = "Copyright 2021 Nándor Mátravölgyi"
__author_email__ = "nandor.matra@gmail.com"
__version__ = "0.1.0"

from .impl import TTLDict



import time
import pytest

from ttldict2 import TTLDict

T = 0.1


def test_ttldict():
    def values(d_):
        return [ttl_value.value for ttl_value in d_.values()]

    with pytest.raises(ValueError):
        TTLDict()
    with pytest.raises(ValueError):
        TTLDict(max_items=0)
    with pytest.raises(ValueError):
        TTLDict(ttl_seconds=0)

    d = TTLDict(max_items=12, ttl_seconds=T * 2.0)

    time.sleep(T * 0.5)
    for i in range(0, 5):
        d[i] = 1 + i
    assert values(d) == [1, 2, 3, 4, 5]

    time.sleep(T * 1)
    for i in range(5, 10):
        d[i] = 1 + i
    assert values(d) == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    for i in range(0, 10):
        assert d[i] == 1 + i
    assert values(d) == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    time.sleep(T * 1)
    assert values(d) == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]  # eat_old has not been called
    assert d.get(-1) is None
    assert d.get(-1, -1) == -1
    assert d.get(-1, d) is d
    assert values(d) == [6, 7, 8, 9, 10]
    for i in range(0, 5):
        with pytest.raises(KeyError):
            _ = d[i]
        assert d.get(i) is None
        assert d.get(i, -1) == -1
        assert d.get(i, d) is d
    assert values(d) == [6, 7, 8, 9, 10]
    for i in range(5, 10):
        assert d[i] == 1 + i
    assert values(d) == [6, 7, 8, 9, 10]

    time.sleep(T * 1.5)
    assert values(d) == [6, 7, 8, 9, 10]
    assert d.get(-1) is None
    assert d.get(-1, -1) == -1
    assert d.get(-1, d) is d
    assert values(d) == []
    for i in range(0, 10):
        with pytest.raises(KeyError):
            _ = d[i]
        assert d.get(i) is None
        assert d.get(i, -1) == -1
        assert d.get(i, d) is d
    assert values(d) == []

    for i in range(0, 5):
        with pytest.raises(KeyError):
            d.pop(i)
        with pytest.raises(KeyError):
            del d[i]
        assert d.pop(i, None) is None
    assert values(d) == []

    for i in range(0, 5):
        assert values(d) == []
        d[i] = 1 + i
        assert values(d) == [1 + i]
        assert d.pop(i) == 1 + i
        assert values(d) == []
        assert i not in d
        assert values(d) == []

    for i in range(0, 5):
        assert values(d) == []
        d[i] = 1 + i
        assert values(d) == [1 + i]
        del d[i]
        assert values(d) == []
        assert i not in d
        assert values(d) == []

    # touch
    for i in range(0, 5):
        d[i] = 1 + i
    assert values(d) == [1, 2, 3, 4, 5]
    time.sleep(T * 1)
    for i in range(0, 5):
        d.get(i, touch=bool(i % 2))
    assert d.get(10, touch=True) is None
    assert values(d) == [1, 3, 5, 2, 4]
    time.sleep(T * 1.5)
    assert values(d) == [1, 3, 5, 2, 4]
    assert d.get(-1) is None
    assert values(d) == [2, 4]
    time.sleep(T * 2.5)
    assert values(d) == [2, 4]
    assert d.get(-1) is None
    assert values(d) == []

    # test ttl drop after update
    d[5] = 7
    time.sleep(T * 1.5)
    d[5] = 8
    time.sleep(T)
    d.drop_old_items()
    time.sleep(T * 2)
    assert d.get(5) is None
    assert values(d) == []

    for i in range(0, 15):
        d[i] = 1 + i
    assert values(d) == [4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
    d[7] = 8
    assert values(d) == [4, 5, 6, 7, 9, 10, 11, 12, 13, 14, 15, 8]



import ast

from setuptools import setup

author = author_email = version = None
with open("ttldict2/__init__.py", encoding="utf-8") as f:
    for line in f:
        if line.startswith("__author__ = "):
            author = ast.literal_eval(line[len("__author__ = ") :])
        elif line.startswith("__author_email__ = "):
            author_email = ast.literal_eval(line[len("__author_email__ = ") :])
        elif line.startswith("__version__ = "):
            version = ast.literal_eval(line[len("__version__ = ") :])


description = "Another dictionary with expiring keys."
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="ttldict2",
    version=version,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Traktormaster/ttldict2",
    author=author,
    author_email=author_email,
    license="Apache 2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
    ],
    packages=["ttldict2"],
)



"""
:copyright: 2021 Nándor Mátravölgyi
:license: Apache2, see LICENSE for more details.
"""
import time
from typing import Optional


class _TTLDictEntry(object):
    """
    Minimal memory footprint item for storing the expiration time.
    """

    __slots__ = ("value", "expires")

    def __init__(self, value, expires):
        self.value = value
        self.expires = expires


class TTLDict(dict):
    """
    Simple dictionary with expiring items of the same life-length. Old items are removed implicitly when
    accessing the dictionary through the supported API.
    This implementation builds on the fact that the builtin dict preserves insertion order. (python 3.7+)

    Only the most relevant subset of the dictionary API is wrapped with the TTL logic. Iterator methods can't
    be wrapped easily for example, because the TTL logic needs to mutate the dict during the iteration. They
    are left usable as-is.

    Not thread-safe!
    """

    __slots__ = ("max_items", "ttl_seconds")

    def __init__(self, max_items: Optional[int] = None, ttl_seconds: Optional[float] = None):
        if max_items is None and ttl_seconds is None:
            raise ValueError("At least one of max_items or ttl_seconds must be specified for TTLDict")
        if max_items is not None and (not isinstance(max_items, int) or max_items < 1):
            raise ValueError("The value %s is invalid as max_items for TTLDict" % (max_items,))
        if ttl_seconds is not None and (not isinstance(ttl_seconds, float) or ttl_seconds <= 0.0):
            raise ValueError("The value %s is invalid as ttl_seconds for TTLDict" % (ttl_seconds,))
        dict.__init__(self)
        self.max_items = max_items
        self.ttl_seconds = ttl_seconds

    def drop_old_items(self):
        ttl_now = time.perf_counter()
        item_view = self.items()
        while item_view:
            key, ttl_value = item_view.__iter__().__next__()
            if ttl_value.expires >= ttl_now:
                break
            dict.__delitem__(self, key)

    def __setitem__(self, key, value):
        ttl_value = dict.pop(self, key, None)
        if ttl_value is None:
            ttl_value = _TTLDictEntry(value, time.perf_counter() + self.ttl_seconds)
            if self.max_items is not None:
                cull = len(self) - self.max_items
                if cull >= 0:
                    key_iter = iter(self)
                    for old_key in [next(key_iter) for _ in range(1 + cull)]:
                        dict.__delitem__(self, old_key)
        else:
            ttl_value.value = value
            ttl_value.expires = time.perf_counter() + self.ttl_seconds
        dict.__setitem__(self, key, ttl_value)

    def __getitem__(self, key):
        self.drop_old_items()
        return dict.__getitem__(self, key).value

    def pop(self, key, *default):
        self.drop_old_items()
        try:
            return dict.pop(self, key).value
        except KeyError:
            if default:
                return default[0]
            raise

    def get(self, key, default=None, touch=False):
        self.drop_old_items()
        if touch:
            ttl_value = dict.pop(self, key, None)
            if ttl_value is None:
                return default
            ttl_value.expires = time.perf_counter() + self.ttl_seconds
            dict.__setitem__(self, key, ttl_value)
            return ttl_value.value
        try:
            return dict.__getitem__(self, key).value
        except KeyError:
            return default
